import os
import click
import datetime
import logging
import requests
from datetime import date, timedelta
from concurrent.futures import ThreadPoolExecutor


from aynse import nse
from aynse import holidays
from aynse.rbi import RBI



logging.basicConfig(level=logging.DEBUG)

@click.group()
def cli():
    """ This is a command line tool to download stock market data to csv files.
        
    """  

def bhavcopy_wrapper(bhavcopy_function, dt, dest):
    try:
        bhavcopy_function(dt, dest)
        return True
    except:
        return False
    


@cli.command("bhavcopy")
@click.option("--dest", "-d", help="Destination directory path", required=True, type=click.Path(exists=True, file_okay=False, dir_okay=True))
@click.option("--from", "-f", "from_", help="From date", type=click.DateTime(["%Y-%m-%d"])) 
@click.option("--to", "-t", help="To date", type=click.DateTime(["%Y-%m-%d"]))
@click.option("--fo/--no-fo", help="Downloads F&O bhavcopy", default=False, type=bool)
@click.option("--idx/--no-idx", help="Downloads Index bhavcopy", default=False, type=bool)
@click.option("--full/--no-full", help="Full Bhavcopy", default=False, type=bool)
def bhavcopy(from_, to, dest, fo, idx, full):
    """Downloads bhavcopy from NSE's website
        
        Download today's bhavcopy
        
        $ jdata bhavcopy -d /path/to/dir

        Download bhavcopy for a date
        
        $ jdata bhavcopy -d /path/to/dir -f 2020-01-01

        Downlad bhavcopy for a date range

        $ jdata bhavcopy -d /path/to/dir -f 2020-01-01 -t 2020-02-01
        
    """ 
        
    downloader = nse.bhavcopy_save
    if full:
        downloader = nse.full_bhavcopy_save
    if idx:
        downloader = nse.bhavcopy_index_save
    if fo:
        downloader = nse.bhavcopy_fo_save
        
    if not from_:
        dt = date.today()
        try:
            path = downloader(dt, dest)
            click.echo("Saved to : " + path)  
        except requests.exceptions.ReadTimeout:
            click.echo("""Error: Timeout while downloading, This may be due to-
        \b1. Bad internet connection
        \b2. Today is holiday or file is not ready yet""", err=True)
    
    if from_ and not to:
        # if from_ provided but not to
        dt = from_.date()
        try:
            path = downloader(dt, dest)
            click.echo("Saved to : " + path)  
        except requests.exceptions.ReadTimeout:
            click.echo("""Error: Timeout while downloading, This may be due to-
        \b1. Bad internet connection
        \b2. {} is holiday or file is not ready yet""".format(dt), err=True)
    
    if from_ and to:
        failed_downloads = []
        date_range = []
        delta = to - from_
        for i in range(delta.days + 1):
            dt = from_ + timedelta(days=i)
            w = dt.weekday()
            if w not in [5,6]:
                date_range.append(dt.date())
        
        with ThreadPoolExecutor() as executor:
            futures = [executor.submit(bhavcopy_wrapper, downloader, dt, dest) for dt in date_range]
            
            with click.progressbar(futures, label="Downloading Bhavcopies") as bar:
                for i, future in enumerate(bar):
                    result = future.result()
                    if not result:
                        failed_downloads.append(date_range[i])
        
             
        """
            for dt in bar:
                try:
                    nse.bhavcopy_save(dt, dest)    
                except requests.exceptions.ReadTimeout:
                    failed_downloads.append(dt)
        """
        click.echo("Saved to : " + dest)
        if failed_downloads:
            click.echo("Failed to download for below dates, these might be holidays, please check -") 
            for dt in failed_downloads:
                click.echo(dt)
        
@cli.command("stock")
@click.option("--symbol", "-s", required=True, help="Stock symbol")
@click.option("--from", "-f", "from_", required=True, help="From date - yyyy-mm-dd")
@click.option("--to", "-t", required=True, help="From date - yyyy-mm-dd")
@click.option("--series", "-S", default="EQ", show_default=True, help="Series - EQ, BE etc.")
@click.option("--output", "-o", default="", help="Full path for output file")
def stock(symbol, from_, to, series, output):
    """Download historical stock data 
    

    $jdata stock --symbol STOCK1 -f yyyy-mm-dd -t yyyy-mm-dd -o file_name.csv
    """
    import traceback
    from_date = datetime.strptime(from_, "%Y-%m-%d").date()
    to_date = datetime.strptime(to, "%Y-%m-%d").date()
    try:
        o = nse.stock_csv(symbol, from_date, to_date, series, output, show_progress=True)
    except Exception as e:
        print(e)
        traceback.print_exc()
    click.echo("\nSaved file to : {}".format(o))
    
@cli.command("index")
@click.option("--symbol", "-s", required=True, help="Stock symbol")
@click.option("--from", "-f", "from_", required=True, help="From date - yyyy-mm-dd")
@click.option("--to", "-t", required=True, help="From date - yyyy-mm-dd")
@click.option("--output", "-o", default="", help="Full path for output file")
def index(symbol, from_, to, output):
    """Download historical index data 
    

    $jdata index --symbol "NIFTY 50" -f yyyy-mm-dd -t yyyy-mm-dd -o file_name.csv
    """
    import traceback
    from_date = datetime.strptime(from_, "%Y-%m-%d").date()
    to_date = datetime.strptime(to, "%Y-%m-%d").date()
    try:
        o = nse.index_csv(symbol, from_date, to_date, output, show_progress=True)
    except Exception as e:
        print(e)
        traceback.print_exc()
    click.echo("\nSaved file to : {}".format(o))
    
@cli.command("derivatives")
@click.option("--symbol", "-s", required=True, help="Stock/Index symbol")
@click.option("--from", "-f", "from_", required=True, help="From date - yyyy-mm-dd")
@click.option("--to", "-t", required=True, help="To date - yyyy-mm-dd")
@click.option("--expiry", "-e", required=True, help="Expiry date - yyyy-mm-dd")
@click.option("--instru", "-i", required=True, help="""FUTSTK - Stock futures, FUTIDX - Index Futures,\tOPTSTK - Stock Options, OPTIDX - Index Options""")
@click.option("--price", "-p", help="Strike price (Only for OPTSTK and OPTIDX)")
@click.option("--ce/--pe", help="--ce for call and --pe for put (Only for OPTSTK and OPTIDX)")
@click.option("--output", "-o", default="", help="Full path of output file")
def stock(symbol, from_, to, expiry, instru, price, ce, output):
    """Sample usage-

    Download stock futures-
    
    \b 
    jdata derivatives -s SBIN -f 2020-01-01 -t 2020-01-30 -e 2020-01-30 -i FUTSTK -o file_name.csv
    
    Download index futures-

    \b
    jdata derivatives -s NIFTY -f 2020-01-01 -t 2020-01-30 -e 2020-01-30 -i FUTIDX -o file_name.csv

    Download stock options-

    \b
    jdata derivatives -s SBIN -f 2020-01-01 -t 2020-01-30 -e 2020-01-30 -i OPTSTK -p 330 --ce -o file_name.csv

    Download index options-

    \b
    jdata derivatives -s NIFTY -f 2020-01-01 -t 2020-01-30 -e 2020-01-23 -i OPTIDX -p 11000 --pe -o file_name.csv


    """
    import traceback
    import sys
    from_date = datetime.strptime(from_, "%Y-%m-%d").date()
    to_date = datetime.strptime(to, "%Y-%m-%d").date()
    expiry = datetime.strptime(expiry, "%Y-%m-%d").date()
    if "OPT" in instru:
        if ce:
            ot = "CE"
        else:
            ot = "PE"
        
        if price:
            price = float(price) 
    else:
        ot = None
    o = nse.derivatives_csv(symbol, from_date, to_date, expiry, instru, price, ot, 
                                output, show_progress=True)

    click.echo("\nSaved file to : {}".format(o))
 
    
    

if __name__ == "__main__":
    cli()





import os
import collections
import json
import pickle
import time
from datetime import datetime, timedelta, date
from concurrent.futures import ThreadPoolExecutor
import click
from appdirs import user_cache_dir

import calendar

import math

try:
    import numpy as np
except:
    np = None

def np_exception(function):
    def wrapper(*args, **kwargs):
        if not np:
            raise ModuleNotFoundError("Please install pandas and numpy using \n pip install pandas")
        return function(*args, **kwargs)

    return wrapper

@np_exception
def np_float(num):
    try:
        return np.float64(num)
    except:
        return np.nan

@np_exception
def np_date(dt):
    try:
        return np.datetime64(dt)
    except:
        pass

    try:
        dt = datetime.strptime(dt, "%d-%b-%Y").date()
        return np.datetime64(dt)
    except:
        pass

    try:
        dt = datetime.strptime(dt, "%d %b %Y").date()
        return np.datetime64(dt)
    except:
        pass



    return np.datetime64('nat') 

    
@np_exception
def np_int(num):
    try:
        return np.int64(num)
    except:
        return 0

def break_dates(from_date, to_date):
    if from_date.replace(day=1) == to_date.replace(day=1):
        return [(from_date, to_date)]
    date_ranges = []
    month_start = from_date
    month_end = month_start.replace(day=calendar.monthrange(month_start.year, from_date.month)[1])
    while(month_end < to_date):
        date_ranges.append((month_start, month_end))
        month_start = month_end + timedelta(days=1)
        month_end = month_start.replace(day=calendar.monthrange(month_start.year, month_start.month)[1])
        if month_end >= to_date:
            date_ranges.append((month_start, to_date))
    return date_ranges


def kw_to_fname(**kw):
    name = "-".join([str(kw[k]) for k in sorted(kw) if k != "self"])
    return name



def cached(app_name):
    """
        Note to self:
            This is a russian doll
            wrapper - actual caching mechanism
            _cached - actual decorator
            cached - wrapper around decorator to make 'app_name' dynamic
    """
    def _cached(function):
        def wrapper(*args, **kw):
            kw.update(zip(function.__code__.co_varnames, args))
            env_dir = os.environ.get("J_CACHE_DIR")
            if not env_dir:
                cache_dir = user_cache_dir(app_name, app_name)
            else:
                cache_dir = os.path.join(env_dir, app_name)

            file_name = kw_to_fname(**kw)
            path = os.path.join(cache_dir, file_name)
            if not os.path.isfile(path):    
                if not os.path.exists(cache_dir):
                    os.makedirs(cache_dir)
                j = function(**kw)
                with open(path, 'wb') as fp:
                    pickle.dump(j, fp)        
            else:
                with open(path, 'rb') as fp:
                    j = pickle.load(fp)
            return j
        return wrapper
    return _cached


def pool(function, params, use_threads=True, max_workers=2):
    if use_threads:
        with ThreadPoolExecutor(max_workers=max_workers) as ex:
            dfs = ex.map(function, *zip(*params))
    else:
        dfs = []
        for param in params:
            try:
                r = function(*param)
            except:
                raise 
            dfs.append(r)
    return dfs

def live_cache(app_name):
    """Caches the output for time_out specified. This is done in order to
    prevent hitting live quote requests to NSE too frequently. This wrapper
    will fetch the quote/live result first time and return the same result for
    any calls within 'time_out' seconds.

    Logic:
        key = concat of args
        try:
            cached_value = self._cache[key]
            if now - self._cache['tstamp'] < time_out
                return cached_value['value']
        except AttributeError: # _cache attribute has not been created yet
            self._cache = {}
        finally:
            val = fetch-new-value
            new_value = {'tstamp': now, 'value': val}
            self._cache[key] = new_value
            return val

    """
    def wrapper(self, *args, **kwargs):
        """Wrapper function which calls the function only after the timeout,
        otherwise returns value from the cache.

        """
        # Get key by just concating the list of args and kwargs values and hope
        # that it does not break the code :P 
        inputs =  [str(a) for a in args] + [str(kwargs[k]) for k in kwargs]
        key = app_name.__name__ + '-'.join(inputs)
        now = datetime.now()
        time_out = self.time_out
        try:
            cache_obj = self._cache[key]
            if now - cache_obj['timestamp'] < timedelta(seconds=time_out):
                return cache_obj['value']
        except:
            self._cache = {}
        value = app_name(self, *args, **kwargs)
        self._cache[key] = {'value': value, 'timestamp': now}
        return value

    return wrapper 




import requests
from bs4 import BeautifulSoup
import pandas as pd
from io import StringIO

def policy_rate_archive(n=10):
    """
    Fetches historical policy rates from RBI website.
    Args:
        n (int): Number of past rates to fetch.
    Returns:
        list: A list of dictionaries with policy rates data.
    """
    base_url = "https://website.rbi.org.in/web/rbi/policy-rate-archive"
    params = {
        "p_p_id": "com_rbi_policy_rate_archive_RBIPolicyRateArchivePortlet_INSTANCE_uwbl",
        "p_p_lifecycle": "0",
        "p_p_state": "normal",
        "p_p_mode": "view",
        "_com_rbi_policy_rate_archive_RBIPolicyRateArchivePortlet_INSTANCE_uwbl_cur": "1",
        "_com_rbi_policy_rate_archive_RBIPolicyRateArchivePortlet_INSTANCE_uwbl_resetCur": "false",
        "_com_rbi_policy_rate_archive_RBIPolicyRateArchivePortlet_INSTANCE_uwbl_delta": str(n),
    }
    
    s = requests.Session()
    r = s.get(base_url, params=params)
    r.raise_for_status()
    
    soup = BeautifulSoup(r.content, "html.parser")
    table_div = soup.find("div", class_="table-responsive")
    
    if not table_div:
        return []

    tables = table_div.find_all("table")
    if not tables:
        return []

    # Assuming the first table is what we need
    html_string = str(tables[0])
    
    # Use StringIO to treat the string as a file for pandas
    df = pd.read_html(StringIO(html_string))[0]

    # Drop rows with all NaN values which can sometimes be parsed from the table
    df = df.dropna(how='all')

    if isinstance(df.columns, pd.MultiIndex):
        df.columns = ['_'.join(map(str, col)).strip() for col in df.columns.values]
    else:
        df.columns = [str(col).strip() for col in df.columns]

    df = df.rename(columns=lambda x: x.lower().replace(' ', '_').replace('unnamed:_0_level_0_', ''))
    
    return df.to_dict('records')

